# Generated by Django 2.2.4 on 2019-08-14 14:55

import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="EventType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("verb", models.CharField(max_length=255, null=True, unique=True)),
                ("enabled", models.BooleanField(default=True, verbose_name="enabled")),
            ],
            options={
                "verbose_name": "event type",
                "verbose_name_plural": "event types",
                "ordering": ("verb",),
            },
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "actor_object_id",
                    models.PositiveIntegerField(
                        null=True, verbose_name="actor object id"
                    ),
                ),
                (
                    "verb",
                    models.CharField(max_length=255, null=True, verbose_name="verb"),
                ),
                (
                    "trigger_object_id",
                    models.PositiveIntegerField(
                        blank=True, null=True, verbose_name="trigger object id"
                    ),
                ),
                (
                    "target_object_id",
                    models.PositiveIntegerField(
                        blank=True, null=True, verbose_name="target object id"
                    ),
                ),
                (
                    "notified",
                    models.BooleanField(default=False, verbose_name="notified"),
                ),
                (
                    "actor_content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="actor_actions",
                        to="contenttypes.ContentType",
                        verbose_name="actor content type",
                    ),
                ),
                (
                    "target_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="target_actions",
                        to="contenttypes.ContentType",
                        verbose_name="target content type",
                    ),
                ),
                (
                    "trigger_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="trigger_actions",
                        to="contenttypes.ContentType",
                        verbose_name="trigger content type",
                    ),
                ),
            ],
            options={"verbose_name": "event", "verbose_name_plural": "events"},
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                ("sent", models.BooleanField(default=False, verbose_name="sent")),
                (
                    "received",
                    models.BooleanField(default=False, verbose_name="received"),
                ),
                ("read", models.BooleanField(default=False, verbose_name="read")),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to="snitch.Event",
                        verbose_name="event",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "notification",
                "verbose_name_plural": "notifications",
                "ordering": ("-created",),
                "abstract": False,
                "swappable": "SNITCH_NOTIFICATION_MODEL",
            },
        ),
    ]



from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractBaseUser
from django.contrib.contenttypes.models import ContentType
from django.db import models

User = get_user_model()


class NotificationQuerySet(models.QuerySet):
    def accessible(self, user: AbstractBaseUser) -> "NotificationQuerySet":
        """Gets the notifications accessible by the given user."""
        if not user.is_authenticated:
            return self.none()
        return self.filter(
            receiver_id=user.pk,
            receiver_content_type=ContentType.objects.get_for_model(User),
        )

    def unread(self) -> "NotificationQuerySet":
        """Gets the unread notifications."""
        return self.filter(read=False)



from typing import TYPE_CHECKING, Tuple, Type

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

from snitch.cooldowns import CoolDownManager
from snitch.exceptions import HandlerError
from snitch.helpers import (
    extract_actor_trigger_target,
    get_notification_model,
    send_event_to_user,
)
from snitch.tasks import create_notification_task

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser

    from snitch.backends import AbstractBackend
    from snitch.cooldowns import AbstractCoolDownManager
    from snitch.models import Event, Notification


class EventHandler:
    """Base event backend to generic even types."""

    ephemeral: bool = False
    dispatch_config: dict = {"args": ("actor", "trigger", "target")}
    action_attribute: str = "actor"
    action_type: str | None = None
    action_id: str | None = None
    click_action: str | None = None
    title: str | None = None
    text: str | None = None
    delay: int = 0
    template_email_async: bool = False
    notification_creation_async: bool = False

    notification_backends: list[Type["AbstractBackend"]] = []
    cool_down_manager_class: Type["AbstractCoolDownManager"] | None = CoolDownManager

    @classmethod
    def extract_actor_trigger_target(cls, method: str, *args, **kwargs):
        """Extracts actor, trigger and target from the args and kwargs
        given as parameters. Override to implement a specific extractor.
        """
        if not isinstance(cls.dispatch_config, dict) or (
            "args" not in cls.dispatch_config and "kwargs" not in cls.dispatch_config
        ):
            raise HandlerError(_("The dispatch config is incorrect."))
        return extract_actor_trigger_target(
            config=cls.dispatch_config, args=args, kwargs=kwargs
        )

    def __init__(self, event: "Event", notification: "Notification | None" = None):
        self.event = event
        self.notification = notification
        self.cool_down_manager = (
            self.cool_down_manager_class(event_handler=self)
            if self.cool_down_manager_class
            else None
        )

    def _default_dynamic_text(self) -> str:
        """Makes an event human readable."""
        text = "{} {}".format(str(self.event.actor), self.event.verb)
        if self.event.trigger:
            text = "{} {}".format(text, str(self.event.trigger))
        if self.event.target:
            text = "{} {}".format(text, str(self.event.target))
        return text

    def should_notify(self, receiver: "models.Model") -> bool:
        """Used by the event to create or not the notifications to the audience. If the
        notification is not created, there isn't any notification sent
        (push, email, etc), and there isn't any record in the database."""
        if self.cool_down_manager:
            return self.cool_down_manager.should_notify(receiver=receiver)
        return True

    def should_send(self, receiver: "models.Model") -> bool:
        """Used by the notification to send or not the notification to the user. If
        returns False, the notification is created in the database but not sent.
        """
        if self.cool_down_manager:
            return self.cool_down_manager.should_send(receiver=receiver)
        return True

    def get_text(self) -> str | None:
        """Override to handle different human readable implementations."""
        return self.text or self._default_dynamic_text()

    def get_title(self) -> str | None:
        """Gets the title for the event. To be hooked."""
        return self.title

    def get_action_type(self) -> str | None:
        """Gets the action type depending on the verb. The actor by default, since
        is the only mandatory field.
        """
        try:
            ContentType = apps.get_model("contenttypes.ContentType")
            action = getattr(self.event, self.action_attribute)
            content_type = ContentType.objects.get_for_model(action)
            return f"{content_type.app_label}.{content_type.model}"
        except AttributeError:
            return None

    def get_action_id(self) -> str | None:
        """Gets the action ID depending on the verb. The actor by default, since
        is the only mandatory field.
        """
        try:
            action = getattr(self.event, self.action_attribute)
            return action.pk
        except AttributeError:
            return None

    def get_click_action(self) -> str | None:
        """Gets the click action depending on the verb. To be hooked."""
        return self.click_action

    def get_delay(self) -> int:
        """Returns and in, number of seconds, that corresponds with the time
        the notification should be delayed.
        """
        return self.delay

    def get_language(self, user: "AbstractBaseUser | None") -> str:
        """Gets the locale for the given used. By default, users the LANGUAGE_CODE
        value from settings.
        """
        return settings.LANGUAGE_CODE

    def get_extra_data(self) -> dict:
        """Adds extra meta data to the backend."""
        return {}

    def audience(self) -> "QuerySet":
        """Gets the audience of the event. None by default, to be hooked by the user."""
        User = get_user_model()
        return User.objects.none()

    def notify(self):
        """If the event is not ephemeral, creates a notification fot each user in the
        audience. In other case, only sends the notification, but doesn't save
        into the database.
        """
        if not self.ephemeral:
            # Creates a notification
            ContentType = apps.get_model("contenttypes.ContentType")
            Notification = get_notification_model()
            for receiver in self.audience().iterator():
                if self.should_notify(receiver=receiver):
                    if self.notification_creation_async:
                        create_notification_task.delay(
                            self.event.pk,
                            receiver.id,
                            ContentType.objects.get_for_model(receiver).pk,
                        )
                    else:
                        notification = Notification(event=self.event, receiver=receiver)
                        notification.save()
        else:
            # Only sends the event to the user
            for user in self.audience().iterator():
                send_event_to_user(event=self.event, user=user)


class EventManager:
    """The event manager in the responsible of handling the registration of the
    handlers with the verbs.
    """

    _registry: dict[str, Type["EventHandler"]]
    _verbs: dict

    def __init__(self):
        self._registry = {}
        self._verbs = {}

    def register(
        self, verb: str, handler: Type["EventHandler"], verbose: str | None = None
    ):
        """Register a handler with a verb, and the verbose form of the verb."""
        if not issubclass(handler, EventHandler):
            raise HandlerError(
                _(f"The handler {handler} have to inherit from EventHandler.")
            )
        self._verbs[verb] = verbose if verbose else verb
        self._registry[verb] = handler

    def choices(self) -> Tuple:
        """Gets a tuple of tuples with the registers verbs and its verbose form, to be
        used as choices."""
        return tuple(self._verbs.items())

    def handler_class(self, verb) -> Type[EventHandler]:
        """Returns a class instance of the handler for the given verb."""
        return self._registry.get(verb, EventHandler)

    def handler(
        self, event: "Event", notification: "Notification | None" = None
    ) -> EventHandler:
        """Returns an instance of the handler for the given event."""
        return self.handler_class(event.verb)(event, notification=notification)


# This global object represents the singleton event manager object
manager: EventManager = EventManager()



from celery import shared_task
from django.db.models import F


@shared_task(serializer="json")
def execute_schedule_task(schedule_id: int) -> int | None:
    from snitch.schedules.models import Schedule

    try:
        schedule = Schedule.objects.get(pk=schedule_id)
    except (Schedule.DoesNotExist, ValueError):
        return None
    schedule.run()
    return schedule_id


@shared_task(serializer="json")
def clean_scheduled_tasks() -> None:
    """Task to clean one shot periodic tasks. Note that this can be done better
    using the proper attributes for PeriodicTask.
    """
    from snitch.schedules.models import Schedule

    schedules = Schedule.objects.filter(limit__isnull=False).filter(
        counter__gte=F("limit")
    )
    # Delete one by one to be sure all the signals and logic beyond delete method
    # are executed
    for schedule in schedules.iterator():
        schedule.delete()



class SnitchError(Exception):
    """Exception for snitch."""


class HandlerError(SnitchError):
    """An error configuring the event handler."""


class AlreadyRegistered(SnitchError):
    """Event handlers already register."""



from celery import shared_task
from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMultiAlternatives

from snitch.helpers import get_notification_model


@shared_task(serializer="json")
def create_notification_task(
    event_pk: int, receiver_id: int, receiver_content_type_id: int
) -> int | None:
    ContentType = apps.get_model("contenttypes.ContentType")
    Event = apps.get_model("snitch.Event")
    Notification = get_notification_model()
    try:
        event = Event.objects.get(pk=event_pk)
        receiver_content_type = ContentType.objects.get(pk=receiver_content_type_id)
        receiver = receiver_content_type.get_object_for_this_type(pk=receiver_id)
    except ObjectDoesNotExist:
        return None
    notification = Notification(event=event, receiver=receiver)
    notification.save()
    return notification.pk


@shared_task(serializer="json")
def send_notification_task(notification_pk: int) -> bool | None:
    """A Celery task to send push notifications related with a given Notification
    model."""

    Notification = get_notification_model()

    try:
        notification = Notification.objects.get(pk=notification_pk)
    except Notification.DoesNotExist:
        return False
    notification.send(send_async=False)
    return None


@shared_task(serializer="json")
def send_email_asynchronously(
    subject: str,
    message_txt: str,
    message: str,
    from_email: str,
    to: list | str,
    cc: list | None,
    bcc: list | None,
    reply_to: list | None,
):
    """Sends an email as a asynchronous task."""
    email = EmailMultiAlternatives(
        subject=subject,
        body=message_txt,
        from_email=from_email,
        to=to,
        cc=cc,
        bcc=bcc,
        reply_to=reply_to,
    )
    email.attach_alternative(message, "text/html")
    email.send()
    return True



import re
import warnings

import bleach
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils import translation

from snitch.settings import ENABLED_SEND_NOTIFICATIONS
from snitch.tasks import send_email_asynchronously


class TemplateEmailMessage:
    """An object to handle emails based on templates, with automatic plain
    alternatives.
    """

    default_template_name: str = ""
    default_subject: str = ""
    default_from_email: str = ""
    fake: bool = False
    use_i18n: bool = False

    def __init__(
        self,
        to: str | list,
        subject: str | None = None,
        context: dict | None = None,
        from_email: str | None = None,
        attaches: list | None = None,
        template_name: str | None = None,
        reply_to: str | list | None = None,
        bcc: str | list | None = None,
        cc: str | list | None = None,
    ):
        self.template_name = (
            self.default_template_name if template_name is None else template_name
        )
        if not self.template_name:
            warnings.warn("You have to specify the template name")
        # Ensure these attributes are lists or tuples
        for attr_name in ["to", "reply_to", "cc", "bcc"]:
            attr = eval(attr_name)
            if (
                attr is not None
                and not isinstance(attr, list)
                and not isinstance(attr, tuple)
            ):
                setattr(self, attr_name, [attr])
            else:
                setattr(self, attr_name, attr)
        self.to = to if isinstance(to, (list, tuple)) else [to]
        self.reply_to = reply_to
        self.bcc = bcc if isinstance(bcc, (list, tuple)) else [bcc]
        self.cc = cc if isinstance(cc, (list, tuple)) else [cc]
        self.subject = self.default_subject if subject is None else subject
        self.from_email = self.default_from_email if from_email is None else from_email
        self.attaches = [] if attaches is None else attaches
        self.default_context = {} if context is None else context

    def get_language(self) -> str:
        """Gets the language for the email."""
        return settings.LANGUAGE_CODE

    def get_context(self) -> dict:
        """Hook to customize context."""
        # Add default context
        current_site = Site.objects.get_current()
        self.default_context.update({"site": current_site})
        return self.default_context

    def preview(self) -> str:
        """Renders the message for a preview."""
        context = self.get_context()
        message = render_to_string(self.template_name, context, using="django")
        return message

    def get_message(self) -> str:
        """Gets the message."""
        context = self.get_context()
        message = render_to_string(self.template_name, context, using="django")
        return message

    def get_plain_message(self, message: str | None = None) -> str:
        """Gets a plain version of the message."""
        if message is None:
            message = self.get_message()
        message_plain = re.sub(r"[\t\n\r\f\v]", "", message)
        message_plain = re.sub(
            "<style.*?>.+</style>", "", message_plain
        )  # Special case for style tag
        message_plain = message_plain.replace("</p>", "\n")
        message_plain = message_plain.replace("</h1>", "\n\n")
        message_plain = bleach.clean(message_plain, strip=True)
        return message_plain.strip()

    def async_send(self, message, message_plain):
        if not self.fake:
            send_email_asynchronously.delay(
                self.subject,
                message_plain,
                message,
                self.from_email,
                self.to,
                self.cc,
                self.bcc,
                self.reply_to,
            )
            if self.attaches:
                warnings.warn(
                    "Attaches will not added to the email, use async=False to send "
                    "attaches."
                )

    def sync_send(self, message, message_plain):
        if not self.fake:
            email = EmailMultiAlternatives(
                subject=self.subject,
                body=message_plain,
                bcc=self.bcc,
                cc=self.cc,
                from_email=self.from_email,
                to=self.to,
                reply_to=self.reply_to,
            )
            email.attach_alternative(message, "text/html")
            for attach in self.attaches:
                attach_file_name, attach_content, attach_content_type = attach
                email.attach(attach_file_name, attach_content, attach_content_type)
            email.send()

    def send(self, use_async: bool = True, language: str | None = None):
        """Sends the email at the moment or using a Celery task."""
        if not ENABLED_SEND_NOTIFICATIONS:
            return

        use_async = not self.attaches and use_async
        if self.use_i18n and settings.USE_I18N:
            language = self.get_language()
            translation.activate(language)
        self.subject = "%s" % self.subject
        message = self.get_message()
        message_plain = self.get_plain_message(message)
        if use_async:
            self.async_send(message, message_plain)
        else:
            self.sync_send(message, message_plain)


class AdminsTemplateEmailMessage(TemplateEmailMessage):
    """Emails only for admins."""

    def __init__(
        self,
        subject: str | None = None,
        context: dict | None = None,
        from_email: str | None = None,
    ):
        to: str | list = [a[1] for a in settings.ADMINS]
        super().__init__(to, subject=subject, context=context, from_email=from_email)


class ManagersTemplateEmailMessage(TemplateEmailMessage):
    """Emails only for mangers."""

    def __init__(
        self,
        subject: str | None = None,
        context: dict | None = None,
        from_email: str | None = None,
    ):
        to: str | list = [manager[1] for manager in settings.MANAGERS]
        super().__init__(to, subject=subject, context=context, from_email=from_email)



from django.contrib import admin

from snitch.schedules.models import Schedule


@admin.register(Schedule)
class ScheduleAdmin(admin.ModelAdmin):
    search_fields = [
        "task__name",
        "actor_content_type",
        "trigger_content_type",
        "target_content_type",
    ]
    list_display = [
        "id",
        "actor",
        "verb",
        "trigger",
        "target",
        "counter",
        "limit",
        "created",
    ]
    list_filter = ["verb"]
    autocomplete_fields = ["task"]



# Generated by Django 3.2.2 on 2021-06-19 06:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("snitch", "0002_alter_eventtype_verb"),
    ]

    operations = [
        migrations.AlterField(
            model_name="eventtype",
            name="verb",
            field=models.CharField(max_length=255, null=True, unique=True),
        ),
    ]



from typing import TYPE_CHECKING, Any, Tuple

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils import translation

from snitch.constants import DEFAULT_CONFIG
from snitch.settings import NOTIFICATION_MODEL

if TYPE_CHECKING:
    from snitch.handlers import EventHandler
    from snitch.models import Event


def get_notification_model():
    """Return the Notification model that is active in this project."""
    try:
        return django_apps.get_model(NOTIFICATION_MODEL, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "NOTIFICATION_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "NOTIFICATION_MODEL refers to model '%s' that has not been installed"
            % NOTIFICATION_MODEL
        )


def explicit_dispatch(
    verb: str, config: dict | None = DEFAULT_CONFIG, *args, **kwargs
) -> Any:
    """Helper to explicit dispatch an event without using a decorator."""
    from snitch.decorators import dispatch

    return dispatch(verb=verb, method=False, config=config)(
        lambda *args, **kwargs: None
    )(*args, **kwargs)


def extract_actor_trigger_target(config: dict, args: Tuple, kwargs: dict) -> Tuple:
    """Extracts the actor, trigger and target using the arguments config given from
    the generic arguments args and kwargs.
    """
    actor = trigger = target = None
    if "args" in config:
        arguments_args = config.get("args", tuple())
        try:
            actor = args[arguments_args.index("actor")]
        except (ValueError, IndexError):
            pass
        try:
            trigger = args[arguments_args.index("trigger")]
        except (ValueError, IndexError):
            pass
        try:
            target = args[arguments_args.index("target")]
        except (ValueError, IndexError):
            pass
    if "kwargs" in config:
        arguments_kwargs = config.get("kwargs", dict())
        try:
            actor = kwargs[arguments_kwargs.get("actor")]
        except (ValueError, KeyError):
            pass
        try:
            trigger = kwargs[arguments_kwargs.get("trigger")]
        except (ValueError, KeyError):
            pass
        try:
            target = kwargs[arguments_kwargs.get("target")]
        except (ValueError, KeyError):
            pass
    return actor, trigger, target


def send_event_to_user(event: "Event", user) -> None:
    """Takes the event and sends it to the user using the backend of the event
    handler.
    """
    handler: "EventHandler" = event.handler()
    if handler.should_send(receiver=user):
        # Activate language for translations
        if settings.USE_I18N:
            language = handler.get_language(user)
            translation.activate(language)
        for backend_class in handler.notification_backends:
            backend = backend_class(event=event, user=user)
            backend.send()


def receiver_content_type_choices() -> "models.Q":
    """Get the posible receivers for a notification."""
    User = get_user_model()  # Here to be able to access after the apps are ready
    choices = models.Q(app_label=User._meta.app_label, model=User._meta.model_name)
    try:
        from push_notifications.models import APNSDevice, GCMDevice

        choices |= models.Q(
            app_label=GCMDevice._meta.app_label, model=GCMDevice._meta.model_name
        ) | models.Q(
            app_label=APNSDevice._meta.app_label, model=APNSDevice._meta.model_name
        )
    except ImportError:
        return choices
    return choices



# Generated by Django 2.2.7 on 2019-11-18 15:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("schedules", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="verb",
            field=models.CharField(
                choices=[
                    ("activated", "activated"),
                    ("confirmed", "confirmed"),
                    ("dummy", "dummy"),
                    ("every hour", "every hour"),
                ],
                max_length=255,
                null=True,
                verbose_name="verb",
            ),
        )
    ]



import hashlib
from typing import TYPE_CHECKING, Any, Callable

from django.core.cache import caches
from django.db import models

if TYPE_CHECKING:
    from snitch.handlers import EventHandler


class AbstractCoolDownManager:
    """The cool down manager handles the cool down feature for notifications, to avoid
    sending several notifications to the same user."""

    event_handler: "EventHandler"

    def __init__(self, event_handler: "EventHandler") -> None:
        self.event_handler = event_handler

    def should_notify(self, receiver: models.Model) -> bool:
        """By default, always notify."""
        return True

    def should_send(self, receiver: models.Model) -> bool:
        """By default, always send."""
        return True


class CoolDownManager(AbstractCoolDownManager):
    """This cool down manager uses the default cache from django to handle the number
    of attempts and the cool down time.
    """

    prefix: str = "snitch"
    cache_alias: str
    attempts: int | Callable[["models.Model"], int] | str
    timeout: int | Callable[["models.Model"], int] | str

    def __init__(self, event_handler: "EventHandler") -> None:
        self.event_handler = event_handler
        self.attempts = getattr(event_handler, "cool_down_attempts", 1)
        self.timeout = getattr(event_handler, "cool_down_time", 0)
        self.cache_alias = getattr(event_handler, "cool_down_cache_alias", "default")

    @property
    def _cache(self) -> Any:
        """Gets the cache proxy using the alias."""
        return caches[self.cache_alias]

    def _key(self, receiver: "models.Model", suffix: str = "") -> str:
        """Get the cache key used by the cool down manager."""
        if receiver.pk is None:
            raise AttributeError("The receiver should have a primary key.")
        key = f"{self.prefix}-cool-down-{self.event_handler.event.verb}-{receiver._meta.app_label}-{receiver._meta.model_name}-{receiver.pk}"
        if suffix:
            key = f"{key}-{suffix}"
        # Use hash function to ensure compatibility
        hashed_key = hashlib.sha256(key.encode()).hexdigest()
        return hashed_key

    def _timeout(self, receiver: "models.Model") -> int:
        """Gets the timeout of the cache in seconds, using the time of the cool down.
        If the timeout is a callable, calls the function, passing the receiver.
        """
        if callable(self.timeout):
            return self.timeout(receiver)
        if isinstance(self.timeout, str):
            return getattr(self.event_handler, self.timeout)(receiver)
        return self.timeout

    def _attempts(self, receiver: "models.Model") -> int:
        """Gets the number of attempts, using the time of the cool down. If the
        attempts number is a callable, calls the function, passing the receiver.
        """
        if callable(self.attempts):
            return self.attempts(receiver)
        if isinstance(self.attempts, str):
            return getattr(self.event_handler, self.attempts)(receiver)
        return self.attempts

    def _check_cool_down(self, receiver: models.Model, suffix: str = "") -> bool:
        """Checks the cool down for the receiver."""
        key = self._key(receiver=receiver, suffix=suffix)
        try:
            self._cache.get_or_set(key, 0, self._timeout(receiver=receiver))
            counter = self._cache.incr(key)
        except ValueError:
            counter = 1
        return counter <= self._attempts(receiver=receiver)

    def should_notify(self, receiver: models.Model) -> bool:
        """Uses the default cool down by default."""
        return self._check_cool_down(receiver=receiver, suffix="notify")

    def should_send(self, receiver: models.Model) -> bool:
        """By default, always send."""
        return self._check_cool_down(receiver=receiver, suffix="send")



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SimpleSnitchConfig(AppConfig):
    """Simple AppConfig which does not do automatic discovery."""

    name: str = "snitch"
    verbose_name: str = _("Snitch")


class SnitchConfig(SimpleSnitchConfig):
    """The default AppConfig for admin which does automatic discovery."""

    default = True

    def ready(self):
        super().ready()
        self.module.autodiscover()



# Generated by Django 2.2.7 on 2019-11-15 12:49

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("django_celery_beat", "0011_auto_20190508_0153"),
    ]

    operations = [
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "actor_object_id",
                    models.PositiveIntegerField(
                        null=True, verbose_name="actor object id"
                    ),
                ),
                (
                    "verb",
                    models.CharField(
                        choices=[
                            ("activated", "activated"),
                            ("confirmed", "confirmed"),
                            ("dummy", "dummy"),
                        ],
                        max_length=255,
                        null=True,
                        verbose_name="verb",
                    ),
                ),
                (
                    "trigger_object_id",
                    models.PositiveIntegerField(
                        blank=True, null=True, verbose_name="trigger object id"
                    ),
                ),
                (
                    "target_object_id",
                    models.PositiveIntegerField(
                        blank=True, null=True, verbose_name="target object id"
                    ),
                ),
                (
                    "limit",
                    models.PositiveIntegerField(
                        help_text="Set to null to always launch the event.",
                        null=True,
                        verbose_name="limit",
                    ),
                ),
                (
                    "counter",
                    models.PositiveIntegerField(
                        blank=True,
                        default=0,
                        help_text="Number of times the schedule has been executed.",
                        verbose_name="counter",
                    ),
                ),
                (
                    "every",
                    models.IntegerField(
                        blank=True,
                        help_text="Number of interval periods to wait before running the task again",
                        null=True,
                        validators=[django.core.validators.MinValueValidator(1)],
                        verbose_name="Number of Periods",
                    ),
                ),
                (
                    "period",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("days", "Days"),
                            ("hours", "Hours"),
                            ("minutes", "Minutes"),
                            ("seconds", "Seconds"),
                            ("microseconds", "Microseconds"),
                        ],
                        help_text="The type of period between task runs (Example: days)",
                        max_length=24,
                        null=True,
                        verbose_name="Interval Period",
                    ),
                ),
                (
                    "minute",
                    models.CharField(default="*", max_length=64, verbose_name="minute"),
                ),
                (
                    "hour",
                    models.CharField(default="*", max_length=64, verbose_name="hour"),
                ),
                (
                    "day_of_week",
                    models.CharField(
                        default="*", max_length=64, verbose_name="day of week"
                    ),
                ),
                (
                    "day_of_month",
                    models.CharField(
                        default="*", max_length=64, verbose_name="day of month"
                    ),
                ),
                (
                    "month_of_year",
                    models.CharField(
                        default="*", max_length=64, verbose_name="month of year"
                    ),
                ),
                (
                    "start_time",
                    models.DateTimeField(
                        blank=True,
                        help_text="Datetime when the schedule should begin triggering the task to run",
                        null=True,
                        verbose_name="Start Datetime",
                    ),
                ),
                (
                    "actor_content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="actor_schedules",
                        to="contenttypes.ContentType",
                        verbose_name="actor content type",
                    ),
                ),
                (
                    "target_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="target_schedules",
                        to="contenttypes.ContentType",
                        verbose_name="target content type",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="django_celery_beat.Periodictask",
                    ),
                ),
                (
                    "trigger_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="trigger_schedules",
                        to="contenttypes.ContentType",
                        verbose_name="trigger content type",
                    ),
                ),
            ],
            options={
                "verbose_name": "schedule",
                "verbose_name_plural": "schedules",
                "ordering": ["-created"],
            },
        )
    ]



# Generated by Django 4.1.6 on 2023-02-20 12:41

from django.contrib.auth import get_user_model
from django.core.exceptions import FieldError
from django.db import migrations, models

from snitch.helpers import get_notification_model


def migrate_users_to_receiver(apps, schema_editor):
    """Gets the user field and uses it to complete the receiver."""
    Notification = get_notification_model()
    User = get_user_model()
    ContentType = apps.get_model("contenttypes", "contenttype")
    try:
        Notification.objects.filter(user__isnull=False).update(
            receiver_id=models.F("user__pk"),
            receiver_content_type=ContentType.objects.get_for_model(User),
        )
    except FieldError:
        ...


class Migration(migrations.Migration):
    dependencies = [
        ("snitch", "0005_notification_receiver_content_type_and_more"),
    ]

    operations = [
        migrations.RunPython(migrate_users_to_receiver),
    ]



# Generated by Django 3.2.2 on 2021-06-10 11:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("schedules", "0002_auto_20191118_0940"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="verb",
            field=models.CharField(max_length=255, null=True, verbose_name="verb"),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SnitchSchedulesAppConfig(AppConfig):
    name: str = "snitch.schedules"
    verbose_name: str = _("Schedules")
    default = True



"""Django app made to integrate generic events that create notifications that
can be sent to users using several backends.

By default, it integrates push notifications and email to send the
notifications.
"""
from pathlib import Path

import django
from django.utils.module_loading import autodiscover_modules
from single_source import get_version

from snitch.decorators import dispatch, register
from snitch.handlers import EventHandler, manager
from snitch.helpers import explicit_dispatch, get_notification_model

__all__ = [
    "register",
    "manager",
    "EventHandler",
    "dispatch",
    "explicit_dispatch",
    "get_notification_model",
]
__version__ = get_version(__name__, Path(__file__).parent.parent) or "1.0.0"
__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)


def autodiscover():
    autodiscover_modules("events", register_to=manager)


if django.VERSION < (3, 2):
    default_app_config = "snitch.apps.SnitchConfig"



# Generated by Django 4.1.6 on 2023-02-20 12:40

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import snitch.helpers


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("snitch", "0004_alter_event_id_alter_eventtype_id_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="receiver_content_type",
            field=models.ForeignKey(
                limit_choices_to=snitch.helpers.receiver_content_type_choices,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
                verbose_name="receiver content type",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="receiver_id",
            field=models.PositiveIntegerField(null=True, verbose_name="receiver id"),
        ),
        migrations.AlterField(
            model_name="notification",
            name="user",
            field=models.ForeignKey(
                blank=True,
                help_text="User owner of the notification. It can be the same as de receiver.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="notifications",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snitch',
 'snitch.migrations',
 'snitch.schedules',
 'snitch.schedules.migrations']

package_data = \
{'': ['*'], 'snitch': ['locale/es_ES/LC_MESSAGES/*']}

install_requires = \
['bleach>=4.1.0',
 'celery>=5.0.0,<6.0.0',
 'django-celery-beat>=2.4.0,<3.0.0',
 'django-model-utils>=4.1.1,<5.0.0',
 'django-push-notifications>=2.0.0',
 'django>=3.0.0',
 'single-source>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'django-snitch',
    'version': '3.1.1',
    'description': 'Django app made to integrate generic events that create notifications that can be sent to users using several backends.',
    'long_description': '=============\nDjango Snitch\n=============\n\n.. image:: https://img.shields.io/pypi/v/django-snitch\n    :target: https://pypi.org/project/django-snitch/\n    :alt: PyPI\n\n.. image:: https://codecov.io/gh/marcosgabarda/django-snitch/branch/main/graph/badge.svg?token=YKC608Q1HX \n    :target: https://codecov.io/gh/marcosgabarda/django-snitch\n\n.. image:: https://img.shields.io/badge/code_style-black-000000.svg\n    :target: https://github.com/ambv/black\n\n.. image:: https://readthedocs.org/projects/django-snitch/badge/?version=latest\n    :target: https://django-snitch.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\nDjango app made to integrate generic events that create notifications that\ncan be sent to users using several backends.\n\nBy default, it integrates **push notifications** and **email** to send the\nnotifications.\n\nMade with Python 3 and Django with :heart:.\n\nQuick start\n-----------\n\n**1** Install using pip:\n\n.. code-block:: bash\n\n    pip install django-snitch\n\n**2** Add "snitch" to your INSTALLED_APPS settings like this:\n\n.. code-block:: python\n\n    INSTALLED_APPS += (\'snitch\',)\n\n**3** Create an ``events.py`` file in your app to register the events:\n\n.. code-block:: python\n\n    import snitch\n    from snitch.backends import PushNotificationBackend, EmailNotificationBackend\n\n    ACTIVATED_EVENT = "activated"\n    CONFIRMED_EVENT = "confirmed"\n\n\n    @snitch.register(ACTIVATED_EVENT)\n    class ActivatedHandler(snitch.EventHandler):\n        title = "Activated!"\n\n\n    @snitch.register(CONFIRMED_EVENT)\n    class ConfirmedHandler(snitch.EventHandler):\n        title = "Confirmed!"\n        notification_backends = [PushNotificationBackend, EmailNotificationBackend]\n\n        # Custom configuration for email backend\n        template_email_kwargs = {"template_name": "email.html"}\n        template_email_async = False\n\n        def audience(self):\n            return get_user_model().objects.all()\n\n\n**4** Use ``dispatch`` decorator to dispatch the event when a function is called:\n\n.. code-block:: python\n\n    from django.db import models\n    from django.utils import timezone\n\n    import snitch\n    from snitch.models import AbstractNotification\n    from tests.app.events import ACTIVATED_EVENT, CONFIRMED_EVENT\n\n\n    class Stuff(models.Model):\n        """Simple stuff model with status."""\n\n        IDLE, ACTIVE, CONFIRMED = 0, 1, 2\n        status = models.PositiveIntegerField(default=IDLE)\n        activated_at = models.DateTimeField(null=True, blank=True)\n        confirmed_at = models.DateTimeField(null=True, blank=True)\n\n        @snitch.dispatch(ACTIVATED_EVENT)\n        def activate(self):\n            self.activated_at = timezone.now()\n\n        @snitch.dispatch(CONFIRMED_EVENT)\n        def confirm(self):\n            self.confirmed_at = timezone.now()\n\n\nCustom Notification model\n-------------------------\n\nYou can, in the same way that ``django.contrib.auth.model.User`` works, swap the\nNotification model, to customize it.\n\nIn order to do that, you should create a model that inherits from\n``AbstractNotification``:\n\n.. code-block:: python\n\n    from django.db import models\n\n    from snitch.models import AbstractNotification\n\n\n    class Notification(AbstractNotification):\n        """Custom notification."""\n\n        extra_field = models.BooleanField(default=False)\n\n\nAnd after that, specify it in the settings:\n\n.. code-block:: python\n\n    SNITCH_NOTIFICATION_MODEL = "app.Notification"\n',
    'author': 'Marcos Gabarda',
    'author_email': 'hey@marcosgabarda.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/marcosgabarda/django-snitch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from django.conf import settings

# Needed to build and publish with Flit
# ------------------------------------------------------------------------------
SECRET_KEY = "snitch"

# Specific project configuration
# ------------------------------------------------------------------------------
NOTIFICATION_EAGER = getattr(settings, "SNITCH_NOTIFICATION_EAGER", False)
ENABLED_SEND_NOTIFICATIONS = getattr(
    settings, "SNITCH_ENABLED_SEND_NOTIFICATIONS", True
)
ENABLED_SEND_EMAILS = getattr(settings, "SNITCH_ENABLED_SEND_EMAILS", True)
NOTIFICATION_MODEL = getattr(
    settings, "SNITCH_NOTIFICATION_MODEL", "snitch.Notification"
)



from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User as AuthUser
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import translation
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel

from snitch.handlers import manager
from snitch.helpers import receiver_content_type_choices
from snitch.managers import NotificationQuerySet
from snitch.settings import NOTIFICATION_EAGER

if TYPE_CHECKING:
    from snitch import EventHandler
    from snitch.backends import AbstractBackend

User = get_user_model()


class EventType(models.Model):
    """Explicit model for Event types, represented by the event verb. It's used to
    enable or disable the generation of notifications.
    """

    verb = models.CharField(max_length=255, null=True, unique=True)
    enabled = models.BooleanField(default=True, verbose_name=_("enabled"))

    class Meta:
        verbose_name = _("event type")
        verbose_name_plural = _("event types")
        ordering = ("verb",)

    def __str__(self) -> str:
        return self.verb


class Event(TimeStampedModel):
    """A 'event' is generated when an 'actor' performs 'verb', involving 'action',
    in the 'target'.

     It could be:
        <actor> <verb>
        <actor> <verb> <target>
        <actor> <verb> <trigger> <target>

    Reference: http://activitystrea.ms/specs/atom/1.0/
    """

    actor_content_type = models.ForeignKey(
        ContentType,
        related_name="actor_actions",
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("actor content type"),
    )
    actor_object_id = models.PositiveIntegerField(_("actor object id"), null=True)
    actor = GenericForeignKey("actor_content_type", "actor_object_id")

    verb = models.CharField(_("verb"), max_length=255, null=True)

    trigger_content_type = models.ForeignKey(
        ContentType,
        related_name="trigger_actions",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("trigger content type"),
    )
    trigger_object_id = models.PositiveIntegerField(
        _("trigger object id"), blank=True, null=True
    )
    trigger = GenericForeignKey("trigger_content_type", "trigger_object_id")

    target_content_type = models.ForeignKey(
        ContentType,
        related_name="target_actions",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("target content type"),
    )
    target_object_id = models.PositiveIntegerField(
        _("target object id"), blank=True, null=True
    )
    target = GenericForeignKey("target_content_type", "target_object_id")

    notified = models.BooleanField(_("notified"), default=False)

    class Meta:
        verbose_name = _("event")
        verbose_name_plural = _("events")

    def __str__(self) -> str:
        handler = self.handler()
        return handler.get_text() or "-"

    def handler(
        self, notification: "AbstractNotification | None" = None
    ) -> "EventHandler":
        """Gets the handler for the event."""
        return manager.handler(self, notification=notification)

    def notify(self) -> None:
        """Creates the notifications associated to this action, ."""
        handler = self.handler()
        handler.notify()
        self.notified = True
        self.save()

    def save(self, *args, **kwargs) -> None:
        super().save(*args, **kwargs)
        if not self.notified:
            self.notify()


class AbstractNotification(TimeStampedModel):
    """A notification is sent to an user, and it's always related with an event."""

    event = models.ForeignKey(
        Event,
        verbose_name=_("event"),
        related_name="notifications",
        on_delete=models.CASCADE,
    )
    receiver = GenericForeignKey("receiver_content_type", "receiver_id")
    receiver_content_type = models.ForeignKey(
        ContentType,
        verbose_name=_("receiver content type"),
        on_delete=models.CASCADE,
        null=True,
        limit_choices_to=receiver_content_type_choices,
    )
    receiver_id = models.PositiveIntegerField(_("receiver id"), null=True)
    sent = models.BooleanField(_("sent"), default=False)
    received = models.BooleanField(_("received"), default=False)
    read = models.BooleanField(_("read"), default=False)

    objects = NotificationQuerySet.as_manager()

    class Meta:
        verbose_name = _("notification")
        verbose_name_plural = _("notifications")
        ordering = ("-created",)
        abstract = True

    def __str__(self) -> str:
        return f"'{str(self.event)}' to {str(self.user)}"

    def _task_kwargs(self, handler: "EventHandler") -> dict:
        """Gets the kwargs for celery task, used in apply_async method."""
        kwargs = {}
        # Delay from event handler
        delay = handler.get_delay()
        if delay:
            kwargs["countdown"] = delay
        return kwargs

    @cached_property
    def user(self) -> AuthUser | None:
        """Get the user if the receiver is an user."""
        return (
            self.receiver
            if self.receiver_content_type == ContentType.objects.get_for_model(User)
            else None
        )

    def receiver_class(self):
        """Gets the class of the device."""
        classes = {ContentType.objects.get_for_model(User): User}
        try:
            from push_notifications.models import APNSDevice, GCMDevice

            classes.update(
                {
                    ContentType.objects.get_for_model(GCMDevice): GCMDevice,
                },
                {
                    ContentType.objects.get_for_model(APNSDevice): APNSDevice,
                },
            )
        except ImportError:
            ...
        return classes.get(self.receiver_content_type)

    def handler(self) -> "EventHandler":
        """Gets the handler for the notification."""
        return self.event.handler(notification=self)

    def send(self, send_async: bool = False) -> None:
        """Sends a push notification to the devices of the user."""
        from snitch.tasks import send_notification_task

        handler: "EventHandler" = self.handler()
        if handler.should_send(receiver=self.receiver):
            if send_async:
                send_notification_task.apply_async(
                    (self.pk,), **self._task_kwargs(handler)
                )
            else:
                # Activate language for translations
                if settings.USE_I18N:
                    language = handler.get_language(self.user)
                    translation.activate(language)
                for backend_class in handler.notification_backends:
                    backend: "AbstractBackend" = backend_class(self)
                    backend.send()
                self.sent = True
                self.save()

    def save(self, *args, **kwargs) -> None:
        """Overwrite to sending push notifications when saving."""
        is_insert: bool = self._state.adding
        super().save(*args, **kwargs)
        if is_insert:
            self.send(send_async=not NOTIFICATION_EAGER)


class Notification(AbstractNotification):
    """Initial notification model that can be swappable."""

    class Meta(AbstractNotification.Meta):
        swappable: str = "SNITCH_NOTIFICATION_MODEL"



import warnings

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import F
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel

import snitch.handlers
from snitch.exceptions import SnitchError

try:
    from django_celery_beat.models import (
        PERIOD_CHOICES,
        CrontabSchedule,
        IntervalSchedule,
        PeriodicTask,
    )
except ImportError:
    raise SnitchError(
        "The snitch.schedules app requires the django-celery-beat package."
    )


class Schedule(TimeStampedModel):
    """A schedule event."""

    # Event data
    actor_content_type = models.ForeignKey(
        ContentType,
        related_name="actor_schedules",
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("actor content type"),
    )
    actor_object_id = models.PositiveIntegerField(_("actor object id"), null=True)
    actor = GenericForeignKey("actor_content_type", "actor_object_id")

    verb = models.CharField(_("verb"), max_length=255, null=True)

    trigger_content_type = models.ForeignKey(
        ContentType,
        related_name="trigger_schedules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("trigger content type"),
    )
    trigger_object_id = models.PositiveIntegerField(
        _("trigger object id"), blank=True, null=True
    )
    trigger = GenericForeignKey("trigger_content_type", "trigger_object_id")

    target_content_type = models.ForeignKey(
        ContentType,
        related_name="target_schedules",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("target content type"),
    )
    target_object_id = models.PositiveIntegerField(
        _("target object id"), blank=True, null=True
    )
    target = GenericForeignKey("target_content_type", "target_object_id")

    # Repeat conditions
    limit = models.PositiveIntegerField(
        _("limit"), null=True, help_text="Set to null to always launch the event."
    )
    counter = models.PositiveIntegerField(
        _("counter"),
        default=0,
        help_text=_("Number of times the schedule has been executed."),
        blank=True,
    )

    # Interval values
    every = models.IntegerField(
        null=True,
        blank=True,
        verbose_name=_("Number of Periods"),
        help_text=_(
            "Number of interval periods to wait before " "running the task again"
        ),
        validators=[MinValueValidator(1)],
    )
    period = models.CharField(
        null=True,
        blank=True,
        max_length=24,
        choices=PERIOD_CHOICES,
        verbose_name=_("Interval Period"),
        help_text=_("The type of period between task runs (Example: days)"),
    )

    # Cron values
    minute = models.CharField(_("minute"), max_length=64, default="*")
    hour = models.CharField(_("hour"), max_length=64, default="*")
    day_of_week = models.CharField(_("day of week"), max_length=64, default="*")
    day_of_month = models.CharField(_("day of month"), max_length=64, default="*")
    month_of_year = models.CharField(_("month of year"), max_length=64, default="*")

    # Start datetime
    start_time = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=_("Start Datetime"),
        help_text=_(
            "Datetime when the schedule should begin " "triggering the task to run"
        ),
    )

    # Periodic task
    task = models.ForeignKey(
        "django_celery_beat.PeriodicTask",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        ordering = ["-created"]
        verbose_name = _("schedule")
        verbose_name_plural = _("schedules")

    def is_active(self) -> bool:
        """Check if the schedule event is active."""
        return self.limit is None or self.counter < self.limit

    def periodic_task_name(self) -> str:
        """Unique task name, using to create it in celery."""
        return f"snitch-scheduled-event-task-{self.pk}"

    def creates_periodic_task_extra_kwarg(self) -> dict:
        """Creates the extra arguments for the periodic task."""
        kwargs = {"one_off": self.limit == 1, "start_time": self.start_time}
        if self.every is not None and self.period is not None:
            kwargs["interval"] = IntervalSchedule.objects.create(
                every=self.every, period=self.period
            )
        else:
            kwargs["crontab"] = CrontabSchedule.objects.create(
                minute=self.minute,
                hour=self.hour,
                day_of_week=self.day_of_week,
                day_of_month=self.day_of_month,
                month_of_year=self.month_of_year,
            )
        return kwargs

    def get_or_creates_periodic_task(self) -> "PeriodicTask":
        """Gets or creates the Celery periodic task."""
        name = self.periodic_task_name()
        if not PeriodicTask.objects.filter(name=name).exists():
            task = PeriodicTask.objects.create(
                name=name,
                task="snitch.schedules.tasks.execute_schedule_task",
                **self.creates_periodic_task_extra_kwarg(),
            )
        else:
            task = PeriodicTask.objects.filter(name=name).first()
        return task

    def run(self) -> None:
        """Executes the schedule, dispatching the event."""
        # Dispatch the event using explicit dispatch
        if not self.actor:
            warnings.warn(f"The schedule {self.pk} has been executed without actor.")
        if self.is_active():
            snitch.dispatch(
                verb=self.verb,
                config={
                    "kwargs": {
                        "actor": "actor",
                        "trigger": "trigger",
                        "target": "target",
                    }
                },
            )(lambda *args, **kwargs: None)(
                actor=self.actor, trigger=self.trigger, target=self.target
            )
            # Updates the counter
            Schedule.objects.filter(pk=self.pk).update(counter=F("counter") + 1)

    def update_task(self) -> None:
        """Syncs the periodic tasks from Celery with the schedule data."""
        # Gets or create task
        if self.task:
            # Updates task fields
            self.task.one_of = self.limit == 1
            self.task.start_time = self.start_time
            # Updates interval values
            interval_fields = ["every", "period"]
            if self.task.interval:
                for interval_field in interval_fields:
                    setattr(
                        self.task.interval,
                        interval_field,
                        getattr(self, interval_field),
                    )
                self.task.interval.save()
            # Updates cron values
            cron_fields: list = [
                "minute",
                "hour",
                "day_of_week",
                "day_of_month",
                "month_of_year",
            ]
            if self.task.crontab:
                for cron_field in cron_fields:
                    setattr(
                        self.task.crontab, cron_field, getattr(self, cron_field) or "*"
                    )
                self.task.crontab.save()
            # Updates kwargs
            self.task.kwargs = f'{{"schedule_id": "{self.pk}"}}'
            # Update enabled
            self.task.enabled = self.is_active()
            # Save task
            self.task.save()

    def delete(self, **kwargs):
        # Delete the task on delete
        self.task.delete()
        return super().delete(**kwargs)

    def save(self, **kwargs) -> None:
        self.clean()
        super().save(**kwargs)
        # Creation/update of the task after the save of the model, to ensure
        # we have an ID.
        if not self.task:
            self.task = self.get_or_creates_periodic_task()
            self.update_task()
            self.save()
        else:
            self.update_task()



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from snitch.models import Event, EventType, Notification


def notify_action(modeladmin, request, queryset):
    """Explicit creates notifications for events."""
    for event in queryset:
        event.notify()
    modeladmin.message_user(request, _("Events notified!"))


notify_action.short_description = _("Notify events")  # type: ignore


def send_action(modeladmin, request, queryset):
    """Explicit sends the notifications using the backend."""
    for notification in queryset:
        notification.send(send_async=True)
    modeladmin.message_user(request, _("Notifications sent!"))


send_action.short_description = _("Send notifications")  # type: ignore


@admin.register(EventType)
class EventTypeAdmin(admin.ModelAdmin):
    list_display = ["id", "verb", "enabled"]


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    list_display = ["id", "actor", "verb", "trigger", "target", "notified", "created"]
    list_filter = ["verb", "notified"]
    actions = [notify_action]


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = [
        "id",
        "event",
        "receiver_content_type",
        "receiver_id",
        "read",
        "received",
        "created",
    ]
    list_filter = ["event__verb", "read", "sent", "receiver_content_type"]
    search_fields = ["receiver_id"]
    actions = [send_action]



import logging
from typing import TYPE_CHECKING, Callable, Tuple, Type, Union

from django.contrib.auth import get_user_model
from django.contrib.auth.models import User as AuthUser
from django.db import models

from snitch.emails import TemplateEmailMessage
from snitch.settings import ENABLED_SEND_NOTIFICATIONS

if TYPE_CHECKING:
    from push_notifications.models import APNSDevice, GCMDevice

    from snitch.handlers import EventHandler
    from snitch.models import AbstractNotification, Event


logger = logging.getLogger(__name__)
User = get_user_model()


class AbstractBackend:
    """Abstract backend class for notifications."""

    def __init__(
        self,
        notification: " AbstractNotification | None" = None,
        event: "Event | None" = None,
        user: AuthUser | None = None,
    ):
        assert notification is not None or (
            event is not None and user is not None
        ), "You should provide a notification or an event and an user."

        self.notification: "AbstractNotification | None" = notification
        self.event: "Event | None " = event
        self.user: AuthUser | None = user
        if self.notification:
            self.handler: EventHandler = self.notification.handler()
            self.user = self.notification.user
        elif self.event:
            self.handler = self.event.handler()

    def send(self):
        """A subclass should to implement the send method."""
        raise NotImplementedError


class PushNotificationBackend(AbstractBackend):
    """A backend class to send push notifications depending on the platform."""

    action_type: str | None
    action_id: str | None
    click_action: str | None
    default_batch_sending: bool = True
    batch_sending: bool

    def __init__(self, *args, **kwargs):
        """Adds attributes for the push notification from the handler."""
        super().__init__(*args, **kwargs)
        self.action_type = self.handler.get_action_type()
        self.action_id = self.handler.get_action_id()
        self.click_action = self.handler.get_click_action()
        self.batch_sending = kwargs.get("batch_sending", self.default_batch_sending)

    def extra_data(self) -> dict:
        """Gets the extra data to add to the push, to be hooked if needed. It tries to
        get an initial dict from the handler.
        """
        return self.handler.get_extra_data()

    def get_devices(
        self,
        device_class: Type["GCMDevice"] | Type["APNSDevice"],
    ) -> "models.QuerySet":
        """Gets the devices using the given class."""
        if self.user is not None:
            return device_class.objects.filter(user=self.user)
        if self.notification and self.notification.receiver_class() == device_class:
            return device_class.objects.filter(pk=self.notification.receiver_id)
        return device_class.objects.none()

    def pre_send(
        self,
        device: "GCMDevice | APNSDevice | None" = None,
    ) -> None:
        """Actions previous to build the message and send, like activate translations if
        needed.
        """
        return None

    def post_send(
        self,
        device: "GCMDevice | APNSDevice | None" = None,
    ) -> None:
        """Actions post to sent the message, like deactivate translations if
        needed.
        """
        return None

    def _build_gcm_message(self) -> Tuple[str | None, dict]:
        """Creates the message for GCM."""
        message: str | None = self.handler.get_text()
        extra = {}
        title: str | None = self.handler.get_title()
        if title:
            extra["title"] = title
        if self.action_type:
            extra["action_type"] = self.action_type
        if self.action_id:
            extra["action_id"] = self.action_id
        if self.click_action:
            extra["click_action"] = self.click_action
        if self.notification:
            extra["notification"] = self.notification.pk
        extra_data = self.extra_data()
        if extra_data:
            extra.update(extra_data)
        return message, extra

    def _build_apns_message(self) -> Tuple[str | dict | None, dict]:
        """Creates the message for APNS."""
        text: str | None = self.handler.get_text()
        message: str | dict | None = text
        extra: dict = {}
        title: str | None = self.handler.get_title()
        if title:
            message = {"title": title, "body": text}
        if self.action_type:
            extra["action_type"] = self.action_type
        if self.action_id:
            extra["action_id"] = self.action_id
        if self.click_action:
            extra["click_action"] = self.click_action
        if self.notification:
            extra["notification"] = self.notification.pk
        extra_data = self.extra_data()
        if extra_data:
            extra.update(extra_data)
        return message, extra

    def _send_to_devices(self, devices: "models.QuerySet", building_method: Callable):
        """Sends a batch of pushes."""
        try:
            from push_notifications.apns import APNSError
            from push_notifications.gcm import GCMError
        except ImportError:
            return None
        if self.batch_sending:
            self.pre_send()
            message, extra = building_method()
            try:
                devices.send_message(message=message, extra=extra)
            except GCMError:
                logger.warning("Error sending a batch GCM push message")
            except APNSError:
                logger.warning("Error sending a batch APNS push message")
            self.post_send()
        else:
            for device in devices:
                self.pre_send(device=device)
                message, extra = building_method()
                try:
                    device.send_message(message=message, extra=extra)
                except GCMError:
                    logger.warning("Error sending a single GCM push message")
                except APNSError:
                    logger.warning("Error sending a single APNS push message")
                self.post_send(device=device)

        return None

    def _send_gcm(self):
        """Send to GCM devices."""
        try:
            from push_notifications.models import GCMDevice
        except ImportError:
            return None
        devices = self.get_devices(GCMDevice)
        self._send_to_devices(devices=devices, building_method=self._build_gcm_message)
        return None

    def _send_apns(self):
        """Send to APNS devices."""
        try:
            from push_notifications.models import APNSDevice
        except ImportError:
            return None
        devices = self.get_devices(APNSDevice)
        self._send_to_devices(devices=devices, building_method=self._build_apns_message)
        return None

    def send(self):
        """Send message for each platform."""
        if ENABLED_SEND_NOTIFICATIONS:
            self._send_gcm()
            self._send_apns()


class EmailNotificationBackend(AbstractBackend):
    """Backend for using the email app to send emails."""

    template_email_kwargs_attr: str = "template_email_kwargs"
    get_email_kwargs_attr: str = "get_email_kwargs_attr"
    get_email_extra_context_attr: str = "get_email_extra_context"
    get_email_subject_attr: str = "get_email_subject"

    def __use_async(self) -> bool:
        """Check if the email can use async, False by default, because the notification
        is already sent using a task."""
        return (
            self.handler.template_email_async  # type: ignore
            if hasattr(self.handler, "template_email_async")
            else False
        )

    def extra_context(self) -> dict:
        """Gets extra context to the email if there is a method in the handler."""
        if hasattr(self.handler, self.get_email_extra_context_attr):
            return getattr(self.handler, self.get_email_extra_context_attr)()
        return {}

    def subject(self) -> str | None:
        """Gets subject of the email if there is a method in the handler."""
        if hasattr(self.handler, self.get_email_subject_attr):
            return getattr(self.handler, self.get_email_subject_attr)()
        return None

    def email_kwargs(self) -> dict | None:
        """Dynamically gets the kwargs for TemplateEmailMessage"""
        kwargs = None
        if hasattr(self.handler, self.get_email_kwargs_attr):
            kwargs = getattr(self.handler, self.get_email_kwargs_attr)()
        elif hasattr(self.handler, self.template_email_kwargs_attr):
            kwargs = getattr(self.handler, self.template_email_kwargs_attr)
        return kwargs

    def send(self):
        """Sends the email."""
        if ENABLED_SEND_NOTIFICATIONS:
            # Gets the handler to extract the arguments from template_email_kwargs
            kwargs = self.email_kwargs()
            if kwargs:
                # Gets to email
                email = (
                    getattr(User, "EMAIL_FIELD")
                    if hasattr(User, "EMAIL_FIELD")
                    else None
                )
                if email:
                    email_field_name = getattr(self.user, "EMAIL_FIELD")
                    kwargs.update({"to": getattr(self.user, email_field_name)})
                # Override subject
                subject = self.subject()
                if subject:
                    kwargs["subject"] = subject
                # Context
                context = kwargs.get("context", {})
                # Adds notification or event
                if self.notification:
                    context.update({"notification": self.notification})
                if self.event:
                    context.update({"event": self.event})
                context.update(self.extra_context())
                kwargs.update({"context": context})
                # Sends email
                email = TemplateEmailMessage(**kwargs)
                email.send(use_async=self.__use_async())



from typing import Callable

from snitch.helpers import extract_actor_trigger_target


def register(verb: str, verbose: str | None = None) -> Callable:
    """Decorator to register an event with its handler.

    @events.register("verb", _("verb verbose"))
    class Handler(events.EventHandler):
        pass

    """
    from snitch.handlers import manager

    def _event_handler_wrapper(event_handler_class):
        manager.register(verb, event_handler_class, verbose=verbose)
        return event_handler_class

    return _event_handler_wrapper


def dispatch(verb: str, method: bool = False, config: dict | None = None) -> Callable:
    """Decorator to dispatch an event when a method or function is called.

    The arguments attribute if to configure how to extract the actor, trigger and
    target from the decorated function arguments.

    config = {
        "args": ("actor", "trigger", "target")  # The position determines de type
        "kwargs": {"actor": "<name>", "trigger": "<name>", "target": "<name>"}
    }

    Example:

    @events.dispatch("verb")
    def method(self, trigger, target=None):
        # self is the actor
        pass

    """
    from django.contrib.contenttypes.models import ContentType

    from snitch.handlers import manager
    from snitch.models import Event, EventType

    def _decorator(func: Callable):
        """Decorator itself."""

        def _wrapper_trigger_action(*args, **kwargs):
            """Wrapped function with the decorator."""

            # Calls the function and saves the result
            result = func(*args, **kwargs)

            # Check if verb is enabled or not
            if EventType.objects.filter(verb=verb, enabled=False).exists():
                return result

            # Extract actor, trigger and target
            # If it isn't specified in arguments attribute, use the handler
            if config is None:
                handler_class = manager.handler_class(verb)
                actor, trigger, target = handler_class.extract_actor_trigger_target(
                    method, *args, **kwargs
                )
            # If it's explicit, use the arguments attribute
            else:
                if not isinstance(config, dict) or (
                    "args" not in config and "kwargs" not in config
                ):
                    return result
                actor, trigger, target = extract_actor_trigger_target(
                    config, args, kwargs
                )

            # Creates the event if there is an actor
            if actor:
                event = Event(
                    actor_content_type=ContentType.objects.get_for_model(actor),
                    actor_object_id=actor.pk,
                    verb=verb,
                )
                if trigger and hasattr(trigger, "pk") and trigger.pk is not None:
                    try:
                        event.trigger_content_type = ContentType.objects.get_for_model(
                            trigger
                        )
                        event.trigger_object_id = trigger.pk
                    except ContentType.DoesNotExist:
                        pass
                if target and hasattr(target, "pk") and target.pk is not None:
                    try:
                        event.target_content_type = ContentType.objects.get_for_model(
                            target
                        )
                        event.target_object_id = target.pk
                    except ContentType.DoesNotExist:
                        pass
                event.save()
            return result

        return _wrapper_trigger_action

    return _decorator
